from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data import Dataset, DataLoader
from torchvision.transforms import ToTensor
import json
from itertools import chain
from skimage import io
from skimage.util import img_as_float
import numpy as np
from time import time

global processed_images


def process_image(image_path):
    gray_image = io.imread(image_path, as_gray=True)
    return img_as_float(gray_image)


class HieroglyphDataset(Dataset):
    def __init__(self, image_paths, labels, transform=None):
        self.image_paths = image_paths
        self.labels = labels
        self.transform = transform

    def __len__(self):
        return len(self.image_paths)

    def __getitem__(self, idx):
        image = processed_images[idx]
        label = self.labels[idx]

        if self.transform:
            image = self.transform(image)

        image = torch.tensor(image, dtype=torch.float32).unsqueeze(0)
        label = torch.tensor(np.argmax(label), dtype=torch.long)
        return image, label


class SimpleCNN(nn.Module):
    def __init__(self):
        super(SimpleCNN, self).__init__()
        self.conv1 = nn.Conv2d(1, 16, kernel_size=3, padding=1)
        self.conv2 = nn.Conv2d(16, 32, kernel_size=3, padding=1)
        self.pool = nn.MaxPool2d(2, 2)
        self.fc1 = nn.Linear(32 * 12 * 12, 128)
        self.fc2 = nn.Linear(128, 95)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = x.view(-1, 12 * 12 * 32)
        x = F.relu(self.fc1(x))
        x = self.fc2(x)

        return x


def get_json_data(filename):
    with open(filename, 'r') as f:
        json_labels = json.load(f)
    keys = list(json_labels.keys())
    one_hot_encoded_keys = {}
    image_paths = []
    n = len(keys)
    for i, key in enumerate(keys):
        one_hot_encoded_keys[key] = [0] * n
        one_hot_encoded_keys[key][i] = 1
        image_paths.append(json_labels[key])

    transformed_data = []
    for key, values in json_labels.items():
        for _ in values:
            transformed_data.append(one_hot_encoded_keys[key])

    image_paths = list(chain.from_iterable(image_paths))
    image_paths = [f"./images/{image}.png" for image in image_paths]
    global processed_images
    processed_images = [process_image(path) for path in image_paths]

    return transformed_data, image_paths


onehot_labels, image_paths = get_json_data('./processed_images_grayscale.json')
train_paths, temp_paths, train_labels, temp_labels = train_test_split(image_paths, onehot_labels, test_size=0.4,
                                                                      random_state=42,
                                                                      shuffle=True)
val_paths, test_paths, val_labels, test_labels = train_test_split(temp_paths, temp_labels, test_size=0.5,
                                                                  random_state=42,
                                                                  shuffle=True)

train_dataset = HieroglyphDataset(train_paths, train_labels, transform=None)

val_dataset = HieroglyphDataset(val_paths, val_labels, transform=None)

test_dataset = HieroglyphDataset(test_paths, test_labels, transform=None)

train_loader = DataLoader(train_dataset, batch_size=4, shuffle=True)

val_loader = DataLoader(val_dataset, batch_size=4, shuffle=False)

test_loader = DataLoader(test_dataset, batch_size=4, shuffle=False)

model = SimpleCNN()
criterion = nn.CrossEntropyLoss()
optimizer = optim.Adam(model.parameters(), lr=0.001)

num_epochs = 20
start_all = time()
for epoch in range(num_epochs):
    start = time()
    model.train()
    running_loss = 0.0
    for images, labels in train_loader:
        optimizer.zero_grad()
        outputs = model(images)
        loss = criterion(outputs, labels)
        loss.backward()
        optimizer.step()
        running_loss += loss.item()
    stop = time()
    print(f"Epoch {epoch + 1}, Loss: {running_loss / len(train_loader)}, time: {stop - start}")
    start = time()
    model.eval()
    val_predictions = []
    val_labels_list = []
    with torch.no_grad():
        for images, labels in val_loader:
            outputs = model(images)
            _, predicted = torch.max(outputs, 1)
            val_predictions.extend(predicted.numpy())
            val_labels_list.extend(labels.numpy())

    val_accuracy = accuracy_score(val_labels_list, val_predictions)
    stop = time()
    print(f'Validation Accuracy: {val_accuracy * 100:.2f}%, time: {stop - start}')
stop_all = time()
print(f'Time: {stop_all - start_all}')

start = time()
test_predictions = []
test_labels_list = []
with torch.no_grad():
    for images, labels in test_loader:
        outputs = model(images)
        _, predicted = torch.max(outputs, 1)
        test_predictions.extend(predicted.numpy())
        test_labels_list.extend(labels.numpy())

test_accuracy = accuracy_score(test_labels_list, test_predictions)
stop = time()
print(f'Test Accuracy: {test_accuracy * 100:.2f}%, time: {stop - start}, time: {stop - start_all}')
